@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2021 Derick Rethans                               |
+   | Copyright (c) 2002-2022 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -20,6 +20,7 @@
 #include "zend_exceptions.h"
 
 #include "debugger_private.h"
+#include "lib/log.h"
 #include "lib/var.h"
 
 extern ZEND_DECLARE_MODULE_GLOBALS(xdebug);
@@ -45,9 +46,10 @@ void xdebug_init_debugger_globals(xdebug_debugger_globals_t *xg)
 	xg->context.do_finish    = 0;
 	xg->context.do_connect_to_client = 0;
 
-	xg->remote_connection_enabled = 0;
-	xg->remote_connection_pid     = 0;
-	xg->breakpoints_allowed       = 0;
+	xg->remote_connection_enabled  = 0;
+	xg->remote_connection_pid      = 0;
+	xg->breakpoints_allowed        = 0;
+	xg->suppress_return_value_step = 0;
 
 	/* Capturing output */
 	if (sapi_module.ub_write != xdebug_ub_write) {
@@ -98,9 +100,15 @@ int xdebug_debugger_bailout_if_no_exec_requested(void)
 
 static void register_compiled_variables(void)
 {
-	function_stack_entry *loop_fse = XDEBUG_VECTOR_TAIL(XG_BASE(stack));
+	function_stack_entry *loop_fse;
 	int                   i;
 
+	if (!XG_BASE(stack)) {
+		return;
+	}
+
+	loop_fse = XDEBUG_VECTOR_TAIL(XG_BASE(stack));
+
 	for (i = 0; i < XDEBUG_VECTOR_COUNT(XG_BASE(stack)); i++, loop_fse--) {
 		if (loop_fse->declared_vars) {
 			continue;
@@ -163,9 +171,6 @@ static int xdebug_handle_hit_value(xdebug_brk_info *brk_info)
 
 int xdebug_do_eval(char *eval_string, zval *ret_zval)
 {
-#if PHP_VERSION_ID < 80000
-	int                old_track_errors;
-#endif
 	volatile int       res = 1;
 	zend_execute_data *original_execute_data = EG(current_execute_data);
 	int                original_no_extensions = EG(no_extensions);
@@ -175,13 +180,7 @@ int xdebug_do_eval(char *eval_string, zval *ret_zval)
 	/* Remember error reporting level and track errors */
 	XG_BASE(error_reporting_override) = EG(error_reporting);
 	XG_BASE(error_reporting_overridden) = 1;
-#if PHP_VERSION_ID < 80000
-	old_track_errors = PG(track_errors);
-#endif
 	EG(error_reporting) = 0;
-#if PHP_VERSION_ID < 80000
-	PG(track_errors) = 0;
-#endif
 
 	XG_DBG(context).inhibit_notifications = 1;
 	XG_DBG(breakpoints_allowed) = 0;
@@ -205,10 +204,8 @@ int xdebug_do_eval(char *eval_string, zval *ret_zval)
 	/* Clean up */
 	EG(error_reporting) = XG_BASE(error_reporting_override);
 	XG_BASE(error_reporting_overridden) = 0;
-#if PHP_VERSION_ID < 80000
-	PG(track_errors) = old_track_errors;
-#endif
 	XG_DBG(breakpoints_allowed) = 1;
+	XG_DBG(suppress_return_value_step) = 0;
 	XG_DBG(context).inhibit_notifications = 0;
 
 	EG(current_execute_data) = original_execute_data;
@@ -219,13 +216,67 @@ int xdebug_do_eval(char *eval_string, zval *ret_zval)
 	return res;
 }
 
+int next_condition_met(void)
+{
+	function_stack_entry *fse;
+	int                   level = 0;
+
+	if (!XG_DBG(context).do_next) {
+		return 0;
+	}
+
+	/* Get latest stack level and function number */
+	if (XG_BASE(stack) && XDEBUG_VECTOR_TAIL(XG_BASE(stack))) {
+		fse = XDEBUG_VECTOR_TAIL(XG_BASE(stack));
+		level = fse->level;
+	}
+
+	if (XG_DBG(context).next_level >= level) {
+		return 1;
+	}
+
+	return 0;
+}
+
+int finish_condition_met(int break_at_return_scope)
+{
+	function_stack_entry *fse;
+	int                   level = 0;
+	int                   func_nr = 0;
+
+	if (!XG_DBG(context).do_finish) {
+		return 0;
+	}
+
+	/* Get latest stack level and function number */
+	if (XG_BASE(stack) && XDEBUG_VECTOR_TAIL(XG_BASE(stack))) {
+		fse = XDEBUG_VECTOR_TAIL(XG_BASE(stack));
+		level = fse->level;
+		func_nr = fse->function_nr;
+	}
+
+	if (!break_at_return_scope && level < XG_DBG(context).finish_level) {
+		return 1;
+	}
+
+	if (break_at_return_scope && level <= XG_DBG(context).finish_level) {
+		return 1;
+	}
+
+	if (
+		(level == XG_DBG(context).finish_level) &&
+		(func_nr > XG_DBG(context).finish_func_nr)
+	) {
+		return 1;
+	}
+
+	return 0;
+}
+
 void xdebug_debugger_statement_call(zend_string *filename, int lineno)
 {
 	xdebug_llist_element *le;
 	xdebug_brk_info      *extra_brk_info;
-	function_stack_entry *fse;
-	int                   level = 0;
-	int                   func_nr = 0;
 
 	if (XG_DBG(context).do_connect_to_client) {
 		XG_DBG(context).do_connect_to_client = 0;
@@ -237,39 +288,26 @@ void xdebug_debugger_statement_call(zend_string *filename, int lineno)
 	}
 
 	if (xdebug_is_debug_connection_active()) {
+		XG_DBG(suppress_return_value_step) = 0;
+
 		if (XG_DBG(context).do_break) {
 			xdebug_brk_info *brk_info = XG_DBG(context).pending_breakpoint;
 
 			XG_DBG(context).do_break = 0;
 			XG_DBG(context).pending_breakpoint = NULL;
 
-			if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), filename, lineno, XDEBUG_BREAK, NULL, 0, NULL, brk_info)) {
+			if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), filename, lineno, XDEBUG_BREAK, NULL, 0, NULL, brk_info, NULL)) {
 				xdebug_mark_debug_connection_not_active();
 				return;
 			}
-		}
-
-		/* Get latest stack level and function number */
-		if (XG_BASE(stack) && XDEBUG_VECTOR_TAIL(XG_BASE(stack))) {
-			fse = XDEBUG_VECTOR_TAIL(XG_BASE(stack));
-			level = fse->level;
-			func_nr = fse->function_nr;
-		} else {
-			level = 0;
-			func_nr = 0;
+			return;
 		}
 
 		/* Check for "finish" */
-		if (
-			XG_DBG(context).do_finish &&
-			(
-				(level < XG_DBG(context).finish_level) ||
-				((level == XG_DBG(context).finish_level) && (func_nr > XG_DBG(context).finish_func_nr))
-			)
-		) {
+		if (finish_condition_met(0)) {
 			XG_DBG(context).do_finish = 0;
 
-			if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), filename, lineno, XDEBUG_STEP, NULL, 0, NULL, NULL)) {
+			if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), filename, lineno, XDEBUG_STEP, NULL, 0, NULL, NULL, NULL)) {
 				xdebug_mark_debug_connection_not_active();
 				return;
 			}
@@ -277,10 +315,10 @@ void xdebug_debugger_statement_call(zend_string *filename, int lineno)
 		}
 
 		/* Check for "next" */
-		if (XG_DBG(context).do_next && XG_DBG(context).next_level >= level) {
+		if (next_condition_met()) {
 			XG_DBG(context).do_next = 0;
 
-			if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), filename, lineno, XDEBUG_STEP, NULL, 0, NULL, NULL)) {
+			if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), filename, lineno, XDEBUG_STEP, NULL, 0, NULL, NULL, NULL)) {
 				xdebug_mark_debug_connection_not_active();
 				return;
 			}
@@ -291,7 +329,7 @@ void xdebug_debugger_statement_call(zend_string *filename, int lineno)
 		if (XG_DBG(context).do_step) {
 			XG_DBG(context).do_step = 0;
 
-			if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), filename, lineno, XDEBUG_STEP, NULL, 0, NULL, NULL)) {
+			if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), filename, lineno, XDEBUG_STEP, NULL, 0, NULL, NULL, NULL)) {
 				xdebug_mark_debug_connection_not_active();
 				return;
 			}
@@ -323,7 +361,7 @@ void xdebug_debugger_statement_call(zend_string *filename, int lineno)
 						}
 					}
 					if (break_ok && xdebug_handle_hit_value(extra_brk_info)) {
-						if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), filename, lineno, XDEBUG_BREAK, NULL, 0, NULL, extra_brk_info)) {
+						if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), filename, lineno, XDEBUG_BREAK, NULL, 0, NULL, extra_brk_info, NULL)) {
 							xdebug_mark_debug_connection_not_active();
 							break;
 						}
@@ -335,23 +373,25 @@ void xdebug_debugger_statement_call(zend_string *filename, int lineno)
 	}
 }
 
-#if PHP_VERSION_ID >= 80000
 void xdebug_debugger_throw_exception_hook(zend_object *exception, zval *file, zval *line, zval *code, char *code_str, zval *message)
 {
 	zend_class_entry *exception_ce = exception->ce;
-#else
-void xdebug_debugger_throw_exception_hook(zval *exception, zval *file, zval *line, zval *code, char *code_str, zval *message)
-{
-	zend_class_entry *exception_ce = Z_OBJCE_P(exception);
-#endif
 	xdebug_brk_info *extra_brk_info;
 
+	register_compiled_variables();
+
 	/* Start JIT if requested and not yet enabled */
 	xdebug_debug_init_if_requested_on_error();
 
-	if (xdebug_is_debug_connection_active() && XG_DBG(breakpoints_allowed)) {
+	if (!xdebug_is_debug_connection_active() || !XG_DBG(breakpoints_allowed)) {
+		return;
+	}
+
+	{
 		int exception_breakpoint_found = 0;
 
+		XG_DBG(suppress_return_value_step) = 1;
+
 		/* Check if we have a wild card exception breakpoint */
 		if (xdebug_hash_find(XG_DBG(context).exception_breakpoints, "*", 1, (void *) &extra_brk_info)) {
 			exception_breakpoint_found = 1;
@@ -376,11 +416,13 @@ void xdebug_debugger_throw_exception_hook(zval *exception, zval *file, zval *lin
 			if (
 				!XG_DBG(context).handler->remote_breakpoint(
 					&(XG_DBG(context)), XG_BASE(stack),
-					Z_STR_P(file), Z_LVAL_P(line), XDEBUG_BREAK,
+					zend_get_executed_filename_ex(), zend_get_executed_lineno(),
+					XDEBUG_BREAK,
 					(char*) STR_NAME_VAL(exception_ce->name),
 					code_str ? code_str : ((code && Z_TYPE_P(code) == IS_STRING) ? Z_STRVAL_P(code) : NULL),
 					message ? Z_STRVAL_P(message) : "",
-					extra_brk_info
+					extra_brk_info,
+					NULL
 				)
 			) {
 				xdebug_mark_debug_connection_not_active();
@@ -393,36 +435,40 @@ void xdebug_debugger_error_cb(zend_string *error_filename, int error_lineno, int
 {
 	xdebug_brk_info *extra_brk_info = NULL;
 
+	register_compiled_variables();
+
 	/* Start JIT if requested and not yet enabled */
 	xdebug_debug_init_if_requested_on_error();
 
-	if (xdebug_is_debug_connection_active() && XG_DBG(breakpoints_allowed)) {
-		/* Send notification with warning/notice/error information */
-		if (XG_DBG(context).send_notifications && !XG_DBG(context).inhibit_notifications) {
-			if (!XG_DBG(context).handler->remote_notification(&(XG_DBG(context)), error_filename, error_lineno, type, error_type_str, buffer)) {
+	if (!xdebug_is_debug_connection_active() || !XG_DBG(breakpoints_allowed)) {
+		return;
+	}
+
+	/* Send notification with warning/notice/error information */
+	if (XG_DBG(context).send_notifications && !XG_DBG(context).inhibit_notifications) {
+		if (!XG_DBG(context).handler->remote_notification(&(XG_DBG(context)), error_filename, error_lineno, type, error_type_str, buffer)) {
+			xdebug_mark_debug_connection_not_active();
+		}
+	}
+
+	/* Check for the pseudo exceptions to allow breakpoints on PHP error statuses */
+	if (
+		xdebug_hash_find(XG_DBG(context).exception_breakpoints, error_type_str, strlen(error_type_str), (void *) &extra_brk_info) ||
+		xdebug_hash_find(XG_DBG(context).exception_breakpoints, "*", 1, (void *) &extra_brk_info)
+	) {
+		if (xdebug_handle_hit_value(extra_brk_info)) {
+			char *type_str = xdebug_sprintf("%ld", type);
+
+			if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), error_filename, error_lineno, XDEBUG_BREAK, error_type_str, type_str, buffer, extra_brk_info, NULL)) {
 				xdebug_mark_debug_connection_not_active();
 			}
-		}
 
-		/* Check for the pseudo exceptions to allow breakpoints on PHP error statuses */
-		if (
-			xdebug_hash_find(XG_DBG(context).exception_breakpoints, error_type_str, strlen(error_type_str), (void *) &extra_brk_info) ||
-			xdebug_hash_find(XG_DBG(context).exception_breakpoints, "*", 1, (void *) &extra_brk_info)
-		) {
-			if (xdebug_handle_hit_value(extra_brk_info)) {
-				char *type_str = xdebug_sprintf("%ld", type);
-
-				if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), error_filename, error_lineno, XDEBUG_BREAK, error_type_str, type_str, buffer, extra_brk_info)) {
-					xdebug_mark_debug_connection_not_active();
-				}
-
-				xdfree(type_str);
-			}
+			xdfree(type_str);
 		}
 	}
 }
 
-static int handle_breakpoints(function_stack_entry *fse, int breakpoint_type)
+static int handle_breakpoints(function_stack_entry *fse, int breakpoint_type, zval *return_value)
 {
 	char            *tmp_name = NULL;
 	size_t           tmp_len = 0;
@@ -437,7 +483,7 @@ static int handle_breakpoints(function_stack_entry *fse, int breakpoint_type)
 		snprintf(
 			tmp_name, tmp_len,
 			"%c/%s",
-			breakpoint_type == XDEBUG_BREAKPOINT_TYPE_CALL ? 'C' : 'R',
+			(breakpoint_type & XDEBUG_BREAKPOINT_TYPE_CALL) ? 'C' : 'R',
 			fse->function.function
 		);
 	}
@@ -452,7 +498,7 @@ static int handle_breakpoints(function_stack_entry *fse, int breakpoint_type)
 		snprintf(
 			tmp_name, tmp_len,
 			"%c/%s::%s",
-			breakpoint_type == XDEBUG_BREAKPOINT_TYPE_CALL ? 'C' : 'R',
+			(breakpoint_type & XDEBUG_BREAKPOINT_TYPE_CALL) ? 'C' : 'R',
 			ZSTR_VAL(fse->function.object_class), fse->function.function
 		);
 	}
@@ -464,10 +510,10 @@ static int handle_breakpoints(function_stack_entry *fse, int breakpoint_type)
 	if (xdebug_hash_find(XG_DBG(context).function_breakpoints, tmp_name, tmp_len - 1, (void *) &extra_brk_info)) {
 		/* Yup, breakpoint found, call handler if the breakpoint is not
 		 * disabled AND handle_hit_value is happy */
-		if (!extra_brk_info->disabled && (extra_brk_info->function_break_type == breakpoint_type)) {
+		if (!extra_brk_info->disabled && (extra_brk_info->function_break_type == (breakpoint_type & XDEBUG_BREAKPOINT_TYPES_MASK))) {
 			if (xdebug_handle_hit_value(extra_brk_info)) {
-				if (fse->user_defined == XDEBUG_BUILT_IN || (breakpoint_type == XDEBUG_BREAKPOINT_TYPE_RETURN)) {
-					if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), fse->filename, fse->lineno, XDEBUG_BREAK, NULL, 0, NULL, extra_brk_info)) {
+				if (fse->user_defined == XDEBUG_BUILT_IN || (breakpoint_type & XDEBUG_BREAKPOINT_TYPE_RETURN)) {
+					if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), fse->filename, fse->lineno, XDEBUG_BREAK, NULL, 0, NULL, extra_brk_info, return_value)) {
 						xdfree(tmp_name);
 						return 0;
 					}
@@ -480,13 +526,32 @@ static int handle_breakpoints(function_stack_entry *fse, int breakpoint_type)
 	}
 	xdfree(tmp_name);
 
+	if (
+		(XG_DBG(context).breakpoint_include_return_value) &&
+		(breakpoint_type & XDEBUG_BREAKPOINT_TYPE_RETURN) &&
+		!(XG_DBG(suppress_return_value_step)) &&
+		return_value
+	) {
+		if (XG_DBG(context).do_step) {
+			XG_DBG(context).do_step = 0;
+		} else if (finish_condition_met(1)) {
+			XG_DBG(context).do_finish = 0;
+		} else {
+			return 1;
+		}
+
+		if (!XG_DBG(context).handler->remote_breakpoint(&(XG_DBG(context)), XG_BASE(stack), fse->filename, fse->lineno, XDEBUG_BREAK, NULL, 0, NULL, extra_brk_info, return_value)) {
+			return 0;
+		}
+	}
+
 	return 1;
 }
 
-void xdebug_debugger_handle_breakpoints(function_stack_entry *fse, int breakpoint_type)
+void xdebug_debugger_handle_breakpoints(function_stack_entry *fse, int breakpoint_type, zval *return_value)
 {
 	if (xdebug_is_debug_connection_active() && XG_DBG(breakpoints_allowed)) {
-		if (!handle_breakpoints(fse, breakpoint_type)) {
+		if (!handle_breakpoints(fse, breakpoint_type, return_value)) {
 			xdebug_mark_debug_connection_not_active();
 		}
 	}
@@ -576,7 +641,7 @@ void xdebug_debugger_rinit(void)
 			)
 			&& !SG(headers_sent)
 		) {
-			xdebug_setcookie("XDEBUG_SESSION", sizeof("XDEBUG_SESSION"), (char*) "", 0, 0, "/", 1, NULL, 0, 0, 1, 0);
+			xdebug_setcookie("XDEBUG_SESSION", sizeof("XDEBUG_SESSION") - 1, (char*) "", 0, 0, "/", 1, NULL, 0, 0, 1, 0);
 			XG_DBG(no_exec) = 1;
 		}
 		zend_string_release(stop_no_exec);
@@ -585,6 +650,7 @@ void xdebug_debugger_rinit(void)
 	xdebug_mark_debug_connection_not_active();
 
 	XG_DBG(breakpoints_allowed) = 1;
+	XG_DBG(suppress_return_value_step) = 0;
 	XG_DBG(detached) = 0;
 	XG_DBG(breakable_lines_map) = xdebug_hash_alloc(2048, (xdebug_hash_dtor_t) xdebug_line_list_dtor);
 	XG_DBG(function_count) = 0;
