@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2020 Derick Rethans                               |
+   | Copyright (c) 2002-2021 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,8 +12,6 @@
    | to obtain it through the world-wide-web, please send a note to       |
    | derick@xdebug.org so we can mail you a copy immediately.             |
    +----------------------------------------------------------------------+
-   | Authors: Derick Rethans <derick@xdebug.org>                          |
-   +----------------------------------------------------------------------+
  */
 #include "php_xdebug.h"
 
@@ -125,7 +123,7 @@ void xdebug_log_stack(const char *error_type_str, char *buffer, const char *erro
 			sent_variables--;
 		}
 
-		tmp_name = xdebug_show_fname(fse->function, 0, 0);
+		tmp_name = xdebug_show_fname(fse->function, XDEBUG_SHOW_FNAME_DEFAULT);
 		xdebug_str_add_fmt(&log_buffer, "PHP %3d. %s(", fse->level, tmp_name);
 		xdfree(tmp_name);
 
@@ -396,7 +394,7 @@ void xdebug_append_printable_stack(xdebug_str *str, int html)
 		if (xdebug_is_stack_frame_filtered(XDEBUG_FILTER_STACK, fse)) {
 			continue;
 		}
-		tmp_name = xdebug_show_fname(fse->function, html, 0);
+		tmp_name = xdebug_show_fname(fse->function, html ? XDEBUG_SHOW_FNAME_ALLOW_HTML : XDEBUG_SHOW_FNAME_DEFAULT);
 		if (html) {
 			xdebug_str_add_fmt(str, formats[3], fse->level, XDEBUG_SECONDS_SINCE_START(fse->nanotime), fse->memory, tmp_name);
 		} else {
@@ -619,7 +617,6 @@ static char *xdebug_handle_stack_trace(int type, char *error_type_str, const cha
 	return printable_stack;
 }
 
-#if PHP_VERSION_ID >= 80000
 static void clear_last_error()
 {
 	if (PG(last_error_message)) {
@@ -627,46 +624,29 @@ static void clear_last_error()
 		PG(last_error_message) = NULL;
 	}
 	if (PG(last_error_file)) {
+# if PHP_VERSION_ID >= 80100
+		zend_string_release(PG(last_error_file));
+# else
 		free(PG(last_error_file));
+# endif
 		PG(last_error_file) = NULL;
 	}
 }
-#else
-static void clear_last_error()
-{
-	if (PG(last_error_message)) {
-		char *s = PG(last_error_message);
-		PG(last_error_message) = NULL;
-		free(s);
-	}
-	if (PG(last_error_file)) {
-		char *s = PG(last_error_file);
-		PG(last_error_file) = NULL;
-		free(s);
-	}
-}
-#endif
 
 /* Error callback for formatting stack traces */
-#if PHP_VERSION_ID >= 80000
-void xdebug_develop_error_cb(int orig_type, const char *error_filename, const unsigned int error_lineno, zend_string *message)
+#if PHP_VERSION_ID >= 80100
+void xdebug_develop_error_cb(int orig_type, zend_string *error_filename, const unsigned int error_lineno, zend_string *message)
 {
 #else
-void xdebug_develop_error_cb(int orig_type, const char *error_filename, const unsigned int error_lineno, const char *format, va_list args)
+void xdebug_develop_error_cb(int orig_type, const char *error_filename, const unsigned int error_lineno, zend_string *message)
 {
-	char *buffer;
-	int buffer_len;
 #endif
-
 	char *error_type_str;
 	int display;
 	int type = orig_type & E_ALL;
 	error_handling_t  error_handling;
 	zend_class_entry *exception_class;
 
-#if PHP_VERSION_ID < 80000
-	buffer_len = vspprintf(&buffer, PG(log_errors_max_len), format, args);
-#endif
 
 	error_type_str = xdebug_error_type(type);
 
@@ -675,14 +655,16 @@ void xdebug_develop_error_cb(int orig_type, const char *error_filename, const un
 			/* no check for PG(last_error_file) is needed since it cannot
 			 * be NULL if PG(last_error_message) is not NULL */
 
-#if PHP_VERSION_ID >= 80000
-			if (zend_string_equals(PG(last_error_message), message)
+			if (!zend_string_equals(PG(last_error_message), message) ||
+				(!PG(ignore_repeated_source) && (
+					(PG(last_error_lineno) != (int)error_lineno) ||
+#if PHP_VERSION_ID >= 80100
+					!zend_string_equals(PG(last_error_file), error_filename)
 #else
-			if (strcmp(PG(last_error_message), buffer)
+					strcmp(PG(last_error_file), error_filename) != 0
 #endif
-					|| (!PG(ignore_repeated_source)
-							&& ((PG(last_error_lineno) != (int)error_lineno)
-									|| strcmp(PG(last_error_file), error_filename)))) {
+				))
+			) {
 					display = 1;
 			} else {
 					display = 0;
@@ -691,28 +673,11 @@ void xdebug_develop_error_cb(int orig_type, const char *error_filename, const un
 			display = 1;
 	}
 
-#if PHP_VERSION_ID < 70300
-	/* Store last error message for error_get_last() */
-	if (display) {
-		clear_last_error();
-		if (!error_filename) {
-			error_filename = "Unknown";
-		}
-		PG(last_error_type) = type;
-		PG(last_error_message) = strdup(buffer);
-		PG(last_error_file) = strdup(error_filename);
-		PG(last_error_lineno) = error_lineno;
-	}
-#endif
 	error_handling  = EG(error_handling);
 	exception_class = EG(exception_class);
-#if PHP_VERSION_ID >= 70300
+
 	/* according to error handling mode, throw exception or show it */
 	if (error_handling == EH_THROW) {
-#else
-	/* according to error handling mode, suppress error, throw exception or show it */
-	if (error_handling != EH_NORMAL) {
-#endif
 		switch (type) {
 			case E_ERROR:
 			case E_CORE_ERROR:
@@ -734,42 +699,33 @@ void xdebug_develop_error_cb(int orig_type, const char *error_filename, const un
 				/* throw an exception if we are in EH_THROW mode
 				 * but DO NOT overwrite a pending exception
 				 */
-#if PHP_VERSION_ID >= 70300
 				if (!EG(exception)) {
-#else
-				if (error_handling == EH_THROW && !EG(exception)) {
-#endif
-#if PHP_VERSION_ID >= 80000
 					zend_throw_error_exception(exception_class, message, 0, type);
-#else
-					zend_throw_error_exception(exception_class, buffer, 0, type);
-#endif
 				}
-#if PHP_VERSION_ID < 80000
-				efree(buffer);
-#endif
 				xdfree(error_type_str);
 				return;
 		}
 	}
 
-#if PHP_VERSION_ID >= 70300
 	/* Store last error message for error_get_last() */
 	if (display) {
 		clear_last_error();
 		if (!error_filename) {
+#if PHP_VERSION_ID >= 80100
+			error_filename = zend_string_init(ZEND_STRL("Unknown"), 0);
+#else
 			error_filename = "Unknown";
+#endif
 		}
 		PG(last_error_type) = type;
-# if PHP_VERSION_ID >= 80000
 		PG(last_error_message) = zend_string_copy(message);
+#if PHP_VERSION_ID >= 80100
+		PG(last_error_file) = zend_string_copy(error_filename);
 #else
-		PG(last_error_message) = strdup(buffer);
-#endif
 		PG(last_error_file) = strdup(error_filename);
+#endif
 		PG(last_error_lineno) = error_lineno;
 	}
-#endif
 
 	if ((EG(error_reporting) | XINI_DEV(force_error_reporting)) & type) {
 		/* Log to logger */
@@ -777,17 +733,13 @@ void xdebug_develop_error_cb(int orig_type, const char *error_filename, const un
 
 #ifdef PHP_WIN32
 			if (type==E_CORE_ERROR || type==E_CORE_WARNING) {
-#if PHP_VERSION_ID >= 80000
 				php_syslog(LOG_ALERT, "PHP %s: %s (%s)", error_type_str, ZSTR_VAL(message), GetCommandLine());
-#else
-				MessageBox(NULL, buffer, error_type_str, MB_OK);
-#endif
 			}
 #endif
-#if PHP_VERSION_ID >= 80000
-			xdebug_log_stack(error_type_str, ZSTR_VAL(message), error_filename, error_lineno);
+#if PHP_VERSION_ID >= 80100
+			xdebug_log_stack(error_type_str, ZSTR_VAL(message), ZSTR_VAL(error_filename), error_lineno);
 #else
-			xdebug_log_stack(error_type_str, buffer, error_filename, error_lineno);
+			xdebug_log_stack(error_type_str, ZSTR_VAL(message), error_filename, error_lineno);
 #endif
 			if (XINI_DEV(dump_globals) && !(XINI_DEV(dump_once) && XG_LIB(dumped))) {
 				char *printable_stack = xdebug_get_printable_superglobals(0);
@@ -816,10 +768,10 @@ void xdebug_develop_error_cb(int orig_type, const char *error_filename, const un
 		if ((PG(display_errors) || XINI_DEV(force_display_errors)) && !PG(during_request_startup)) {
 			char *printable_stack;
 
-#if PHP_VERSION_ID >= 80000
-			printable_stack = xdebug_handle_stack_trace(type, error_type_str, error_filename, error_lineno, ZSTR_VAL(message));
+#if PHP_VERSION_ID >= 80100
+			printable_stack = xdebug_handle_stack_trace(type, error_type_str, ZSTR_VAL(error_filename), error_lineno, ZSTR_VAL(message));
 #else
-			printable_stack = xdebug_handle_stack_trace(type, error_type_str, error_filename, error_lineno, buffer);
+			printable_stack = xdebug_handle_stack_trace(type, error_type_str, error_filename, error_lineno, ZSTR_VAL(message));
 #endif
 
 			if (XG_LIB(do_collect_errors) && (type != E_ERROR) && (type != E_COMPILE_ERROR) && (type != E_USER_ERROR)) {
@@ -830,22 +782,22 @@ void xdebug_develop_error_cb(int orig_type, const char *error_filename, const un
 			}
 		} else if (XG_LIB(do_collect_errors)) {
 			char *printable_stack;
-#if PHP_VERSION_ID >= 80000
-			printable_stack = xdebug_get_printable_stack(PG(html_errors), type, ZSTR_VAL(message), error_filename, error_lineno, 1);
+#if PHP_VERSION_ID >= 80100
+			printable_stack = xdebug_get_printable_stack(PG(html_errors), type, ZSTR_VAL(message), ZSTR_VAL(error_filename), error_lineno, 1);
 #else
-			printable_stack = xdebug_get_printable_stack(PG(html_errors), type, buffer, error_filename, error_lineno, 1);
+			printable_stack = xdebug_get_printable_stack(PG(html_errors), type, ZSTR_VAL(message), error_filename, error_lineno, 1);
 #endif
 			xdebug_llist_insert_next(XG_DEV(collected_errors), XDEBUG_LLIST_TAIL(XG_DEV(collected_errors)), printable_stack);
 		}
 	}
 
 	{
-		zend_string *tmp_error_filename = zend_string_init(error_filename, strlen(error_filename), 0);
-#if PHP_VERSION_ID >= 80000
-		xdebug_debugger_error_cb(tmp_error_filename, error_lineno, type, error_type_str, ZSTR_VAL(message));
+#if PHP_VERSION_ID >= 80100
+		zend_string *tmp_error_filename = zend_string_copy(error_filename);
 #else
-		xdebug_debugger_error_cb(tmp_error_filename, error_lineno, type, error_type_str, buffer);
+		zend_string *tmp_error_filename = zend_string_init(error_filename, strlen(error_filename), 0);
 #endif
+		xdebug_debugger_error_cb(tmp_error_filename, error_lineno, type, error_type_str, ZSTR_VAL(message));
 		zend_string_release(tmp_error_filename);
 	}
 
@@ -882,16 +834,9 @@ void xdebug_develop_error_cb(int orig_type, const char *error_filename, const un
 					sapi_header_op(SAPI_HEADER_REPLACE, &ctr);
 				}
 				/* the parser would return 1 (failure), we can bail out nicely */
-#if PHP_VERSION_ID >= 80000
 				if (!(orig_type & E_DONT_BAIL)) {
-#else
-				if (type != E_PARSE) {
-#endif
 					/* restore memory limit */
 					zend_set_memory_limit(PG(memory_limit));
-#if PHP_VERSION_ID < 80000
-					efree(buffer);
-#endif
 					zend_objects_store_mark_destructed(&EG(objects_store));
 					_zend_bailout((char*) __FILE__, __LINE__);
 					return;
@@ -899,42 +844,11 @@ void xdebug_develop_error_cb(int orig_type, const char *error_filename, const un
 			}
 			break;
 	}
-
-#if PHP_VERSION_ID < 70400
-	/* Log if necessary */
-	if (!display) {
-		efree(buffer);
-		return;
-	}
-
-	if (PG(track_errors) && EG(active)) {
-		zval tmp;
-		ZVAL_STRINGL(&tmp, buffer, buffer_len);
-
-		if (EG(current_execute_data)) {
-			if (zend_set_local_var_str("php_errormsg", sizeof("php_errormsg")-1, &tmp, 0) == FAILURE) {
-				zval_ptr_dtor(&tmp);
-			}
-		} else {
-			zend_hash_str_update(&EG(symbol_table), "php_errormsg", sizeof("php_errormsg"), &tmp);
-		}
-	}
-#endif
-
-#if PHP_VERSION_ID < 80000
-	efree(buffer);
-#endif
 }
 
-#if PHP_VERSION_ID >= 80000
 void xdebug_develop_throw_exception_hook(zend_object *exception, zval *file, zval *line, zval *code, char *code_str, zval *message)
 {
 	zend_class_entry *exception_ce = exception->ce;
-#else
-void xdebug_develop_throw_exception_hook(zval *exception, zval *file, zval *line, zval *code, char *code_str, zval *message)
-{
-	zend_class_entry *exception_ce = Z_OBJCE_P(exception);
-#endif
 	zval *xdebug_message_trace, *previous_exception;
 	char *exception_trace;
 	xdebug_str tmp_str = XDEBUG_STR_INITIALIZER;
@@ -942,11 +856,7 @@ void xdebug_develop_throw_exception_hook(zval *exception, zval *file, zval *line
 
 	previous_exception = zend_read_property(exception_ce, exception, "previous", sizeof("previous")-1, 1, &dummy);
 	if (previous_exception && Z_TYPE_P(previous_exception) == IS_OBJECT) {
-#if PHP_VERSION_ID >= 80000
 		xdebug_message_trace = zend_read_property(exception_ce, Z_OBJ_P(previous_exception), "xdebug_message", sizeof("xdebug_message")-1, 1, &dummy);
-#else
-		xdebug_message_trace = zend_read_property(exception_ce, previous_exception, "xdebug_message", sizeof("xdebug_message")-1, 1, &dummy);
-#endif
 		if (xdebug_message_trace && Z_TYPE_P(xdebug_message_trace) != IS_NULL) {
 			xdebug_str_add(&tmp_str, Z_STRVAL_P(xdebug_message_trace), 0);
 		}
@@ -958,6 +868,10 @@ void xdebug_develop_throw_exception_hook(zval *exception, zval *file, zval *line
 	xdebug_append_error_description(&tmp_str, PG(html_errors), STR_NAME_VAL(exception_ce->name), message ? Z_STRVAL_P(message) : "", Z_STRVAL_P(file), Z_LVAL_P(line));
 	xdebug_append_printable_stack(&tmp_str, PG(html_errors));
 	exception_trace = tmp_str.d;
+
+#if PHP_VERSION_ID >= 80200
+	exception_ce->ce_flags |= ZEND_ACC_ALLOW_DYNAMIC_PROPERTIES;
+#endif
 	zend_update_property_string(exception_ce, exception, "xdebug_message", sizeof("xdebug_message")-1, exception_trace);
 
 	if (XG_BASE(last_exception_trace)) {
@@ -985,6 +899,11 @@ void xdebug_develop_throw_exception_hook(zval *exception, zval *file, zval *line
    Returns the stack depth */
 PHP_FUNCTION(xdebug_get_stack_depth)
 {
+	if (!XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP)) {
+		php_error(E_WARNING, "Function must be enabled in php.ini by setting 'xdebug.mode' to 'develop'");
+		RETURN_LONG(0);
+	}
+
 	/* We substract one so that the function call to xdebug_get_stack_depth()
 	 * is not part of the returned depth. */
 	RETURN_LONG(XDEBUG_VECTOR_COUNT(XG_BASE(stack)) - 1);
@@ -999,6 +918,7 @@ PHP_FUNCTION(xdebug_get_function_stack)
 	unsigned int          j;
 	zval                 *frame;
 	zval                 *params;
+	int                   variadic_opened = 0;
 
 	if (!XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP)) {
 		php_error(E_WARNING, "Function must be enabled in php.ini by setting 'xdebug.mode' to 'develop'");
@@ -1031,9 +951,9 @@ PHP_FUNCTION(xdebug_get_function_stack)
 		if (fse->function.function) {
 			add_assoc_string_ex(frame, "function", HASH_KEY_SIZEOF("function"), fse->function.function);
 		}
-		if (fse->function.class_name) {
+		if (fse->function.object_class) {
 			add_assoc_string_ex(frame, "type",     HASH_KEY_SIZEOF("type"),     (char*) (fse->function.type == XFUNC_STATIC_MEMBER ? "static" : "dynamic"));
-			add_assoc_str_ex(frame,    "class",    HASH_KEY_SIZEOF("class"),    zend_string_copy(fse->function.class_name));
+			add_assoc_str_ex(frame,    "class",    HASH_KEY_SIZEOF("class"),    zend_string_copy(fse->function.object_class));
 		}
 		add_assoc_str_ex(frame, "file", HASH_KEY_SIZEOF("file"), zend_string_copy(fse->filename));
 		add_assoc_long_ex(frame, "line", HASH_KEY_SIZEOF("line"), fse->lineno);
@@ -1044,7 +964,6 @@ PHP_FUNCTION(xdebug_get_function_stack)
 		add_assoc_zval_ex(frame, "params", HASH_KEY_SIZEOF("params"), params);
 
 		for (j = 0; j < sent_variables; j++) {
-			int         variadic_opened = 0;
 			xdebug_str *argument = NULL;
 
 			if (fse->var[j].is_variadic) {
@@ -1071,7 +990,7 @@ PHP_FUNCTION(xdebug_get_function_stack)
 			if (fse->var[j].name && !variadic_opened && argument) {
 				add_assoc_stringl_ex(params, ZSTR_VAL(fse->var[j].name), ZSTR_LEN(fse->var[j].name), argument->d, argument->l);
 			} else {
-				add_index_stringl(params, j - 1, argument->d, argument->l);
+				add_index_stringl(params, j - variadic_opened, argument->d, argument->l);
 			}
 			if (argument) {
 				xdebug_str_free(argument);
