@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2020 Derick Rethans                               |
+   | Copyright (c) 2002-2021 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,8 +12,6 @@
    | to obtain it through the world-wide-web, please send a note to       |
    | derick@xdebug.org so we can mail you a copy immediately.             |
    +----------------------------------------------------------------------+
-   | Authors: Derick Rethans <derick@xdebug.org>                          |
-   +----------------------------------------------------------------------+
  */
 
 #ifndef __XDEBUG_LIBRARY_H__
@@ -27,6 +25,8 @@
 #include "zend_API.h"
 #include "compat.h"
 
+extern int xdebug_global_mode;
+
 typedef struct xdebug_var_name {
 	zend_string *name;
 	zval         data;
@@ -45,6 +45,9 @@ typedef struct xdebug_var_name {
 #define XFUNC_REQUIRE        0x13
 #define XFUNC_REQUIRE_ONCE   0x14
 #define XFUNC_MAIN           0x15
+#if PHP_VERSION_ID >= 80100
+# define XFUNC_FIBER         0x16
+#endif
 
 #define XFUNC_ZEND_PASS      0x20
 
@@ -119,7 +122,8 @@ typedef struct xdebug_var_name {
 #define XDEBUG_ERROR_ENCODING_NOT_SUPPORTED        900
 
 typedef struct _xdebug_func {
-	zend_string *class_name;
+	zend_string *object_class;
+	zend_string *scope_class;
 	char *function;
 	int   type;
 	int   internal;
@@ -192,7 +196,6 @@ struct _xdebug_multi_opcode_handler_t
 };
 
 typedef struct _xdebug_library_globals_t {
-	int                    mode;
 	int                    start_with_request; /* One of the XDEBUG_START_WITH_REQUEST_* constants */
 	int                    start_upon_error;   /* One of the XDEBUG_START_UPON_ERROR_* constants */
 	int                    mode_from_environment; /* Keeps track whether the mode was set with XDEBUG_MODE for diagnostics purposes */
@@ -221,12 +224,18 @@ typedef struct _xdebug_library_globals_t {
 } xdebug_library_globals_t;
 
 typedef struct _xdebug_library_settings_t {
+	char         *requested_mode;
+
 	char         *output_dir;
 	char         *trigger_value;
 
 	char         *file_link_format;
 	char         *filename_format;
 
+	/* Whether to use zlib compression for profiling and trace files, if ZLIB support
+	 * is enabled */
+	zend_bool     use_compression;
+
 	/* variable dumping limitation settings */
 	zend_long     display_max_children;
 	zend_long     display_max_data;
@@ -255,10 +264,10 @@ void xdebug_disable_opcache_optimizer(void);
 #define XDEBUG_MODE_GCSTATS      1<<3
 #define XDEBUG_MODE_PROFILING    1<<4
 #define XDEBUG_MODE_TRACING      1<<5
-int xdebug_lib_set_mode(char *mode);
+int xdebug_lib_set_mode(const char *mode);
 
-#define XDEBUG_MODE_IS_OFF() ((XG(globals.library.mode) == XDEBUG_MODE_OFF))
-#define XDEBUG_MODE_IS(v) ((XG(globals.library.mode) & (v)) ? 1 : 0)
+#define XDEBUG_MODE_IS_OFF() ((xdebug_global_mode == XDEBUG_MODE_OFF))
+#define XDEBUG_MODE_IS(v) ((xdebug_global_mode & (v)) ? 1 : 0)
 #define RETURN_IF_MODE_IS_NOT(m) if (!XDEBUG_MODE_IS((m))) { return; }
 #define RETURN_FALSE_IF_MODE_IS_NOT(m) if (!XDEBUG_MODE_IS((m))) { RETURN_FALSE; }
 #define WARN_AND_RETURN_IF_MODE_IS_NOT(m) if (!XDEBUG_MODE_IS((m))) { php_error(E_NOTICE, "Functionality is not enabled"); return; }
@@ -268,11 +277,12 @@ int xdebug_lib_set_mode(char *mode);
 #define XDEBUG_START_WITH_REQUEST_NO          3
 #define XDEBUG_START_WITH_REQUEST_TRIGGER     4
 int xdebug_lib_set_start_with_request(char *value);
-int xdebug_lib_start_with_request(void);
+int xdebug_lib_start_with_request(int for_mode);
 int xdebug_lib_start_with_trigger(int for_mode, char **found_trigger_value);
 int xdebug_lib_start_if_mode_is_trigger(int for_mode);
 int xdebug_lib_never_start_with_request(void);
 int xdebug_lib_get_start_with_request(void);
+int xdebug_lib_has_shared_secret(void);
 
 #define XDEBUG_START_UPON_ERROR_DEFAULT     1
 #define XDEBUG_START_UPON_ERROR_YES         2
@@ -281,6 +291,8 @@ int xdebug_lib_set_start_upon_error(char *value);
 int xdebug_lib_start_upon_error(void);
 int xdebug_lib_get_start_upon_error(void);
 
+const char *xdebug_lib_mode_from_value(int mode);
+
 void xdebug_lib_set_active_data(zend_execute_data *execute_data);
 void xdebug_lib_set_active_stack_entry(function_stack_entry *fse);
 void xdebug_lib_set_active_symbol_table(HashTable *symbol_table);
@@ -306,6 +318,7 @@ int xdebug_call_original_opcode_handler_if_set(int opcode, XDEBUG_OPCODE_HANDLER
 char *xdebug_lib_get_output_dir(void);
 
 void xdebug_llist_string_dtor(void *dummy, void *elem);
+char* xdebug_wrap_location_around_function_name(const char *prefix, zend_op_array *opa, char *fname);
 char* xdebug_wrap_closure_location_around_function_name(zend_op_array *opa, char *fname);
 
 void xdebug_lib_register_compiled_variables(function_stack_entry *fse, zend_op_array *op_array);
